@@ -3,7 +3,6 @@ set -e
 
 meson setup build \
   --werror        \
-  -Dlibplacebo:werror=false \
   -Dc_args="-Wno-error=deprecated -Wno-error=deprecated-declarations" \
   -Db_sanitize=address,undefined \
   -Dcdda=enabled          \
@@ -13,7 +12,6 @@ meson setup build \
   -Dlibmpv=true           \
   -Dmanpage-build=enabled \
   -Dpipewire=enabled      \
-  -Dshaderc=enabled       \
   -Dtests=true            \
   -Dvulkan=enabled
 meson compile -C build
