@@ -56,12 +56,10 @@ struct hwmap_pairs {
 // We cannot discover which pairs of hardware formats need to use hwmap to
 // convert between the formats, so we need a lookup table.
 static const struct hwmap_pairs hwmap_pairs[] = {
-#if HAVE_VULKAN_INTEROP
     {
         .first_fmt = IMGFMT_VAAPI,
         .second_fmt = IMGFMT_VULKAN,
     },
-#endif
     {
         .first_fmt = IMGFMT_DRMPRIME,
         .second_fmt = IMGFMT_VAAPI,
@@ -94,10 +92,12 @@ static bool select_format(struct priv *p, int input_fmt,
     if (!input_fmt)
         return false;
 
-    // If the input format is a hw format, then we shouldn't be doing this
-    // format selection here at all.
+    // If the input format is a hw format, then we won't be doing any sort of
+    // conversion. Just assume that it will pass-through successfully.
     if (IMGFMT_IS_HWACCEL(input_fmt)) {
-        return false;
+        *out_hw_input_fmt = input_fmt;
+        *out_hw_output_fmt = input_fmt;
+        return true;
     }
 
     // If there is no capability to do uploads or conversions during uploads,
@@ -143,7 +143,7 @@ static bool select_format(struct priv *p, int input_fmt,
     return true;
 }
 
-static void process(struct mp_filter *f)
+static void hwupload_process(struct mp_filter *f)
 {
     struct priv *p = f->priv;
 
@@ -236,7 +236,7 @@ error:
     mp_filter_internal_mark_failed(f);
 }
 
-static void destroy(struct mp_filter *f)
+static void hwupload_destroy(struct mp_filter *f)
 {
     struct priv *p = f->priv;
 
@@ -247,8 +247,8 @@ static void destroy(struct mp_filter *f)
 static const struct mp_filter_info hwupload_filter = {
     .name = "hwupload",
     .priv_size = sizeof(struct priv),
-    .process = process,
-    .destroy = destroy,
+    .process = hwupload_process,
+    .destroy = hwupload_destroy,
 };
 
 // The VO layer might have restricted format support. It might actually
