@@ -50,13 +50,13 @@
 #include "common/common.h"
 
 #if HAVE_COCOA
-#include "osdep/macosx_events.h"
+#include "osdep/mac/app_bridge.h"
 #endif
 
 #define input_lock(ictx)    mp_mutex_lock(&ictx->mutex)
 #define input_unlock(ictx)  mp_mutex_unlock(&ictx->mutex)
 
-#define MP_MAX_KEY_DOWN 4
+#define MP_MAX_KEY_DOWN 16
 
 struct cmd_bind {
     int keys[MP_MAX_KEY_DOWN];
@@ -79,8 +79,6 @@ struct cmd_bind_section {
 
 #define MP_MAX_SOURCES 10
 
-#define MAX_ACTIVE_SECTIONS 50
-
 struct active_section {
     char *name;
     int flags;
@@ -142,7 +140,7 @@ struct input_ctx {
     int num_sections;
 
     // List currently active command sections
-    struct active_section active_sections[MAX_ACTIVE_SECTIONS];
+    struct active_section *active_sections;
     int num_active_sections;
 
     unsigned int mouse_event_counter;
@@ -179,6 +177,7 @@ struct input_opts {
     bool vo_key_input;
     bool test;
     bool allow_win_drag;
+    bool preprocess_wheel;
 };
 
 const struct m_sub_options input_config = {
@@ -198,6 +197,7 @@ const struct m_sub_options input_config = {
         {"input-cursor", OPT_BOOL(enable_mouse_movements)},
         {"input-vo-keyboard", OPT_BOOL(vo_key_input)},
         {"input-media-keys", OPT_BOOL(use_media_keys)},
+        {"input-preprocess-wheel", OPT_BOOL(preprocess_wheel)},
 #if HAVE_SDL2_GAMEPAD
         {"input-gamepad", OPT_BOOL(use_gamepad)},
 #endif
@@ -217,6 +217,7 @@ const struct m_sub_options input_config = {
         .builtin_bindings = true,
         .vo_key_input = true,
         .allow_win_drag = true,
+        .preprocess_wheel = true,
     },
     .change_flags = UPDATE_INPUT,
 };
@@ -731,18 +732,23 @@ static void mp_input_feed_key(struct input_ctx *ictx, int code, double scale,
     if (!force_mouse && opts->doubleclick_time && MP_KEY_IS_MOUSE_BTN_DBL(unmod))
         return;
     int units = 1;
-    if (MP_KEY_IS_WHEEL(unmod) && !process_wheel(ictx, unmod, &scale, &units))
+    if (MP_KEY_IS_WHEEL(unmod) && opts->preprocess_wheel && !process_wheel(ictx, unmod, &scale, &units))
         return;
     interpret_key(ictx, code, scale, units);
     if (code & MP_KEY_STATE_DOWN) {
         code &= ~MP_KEY_STATE_DOWN;
         if (ictx->last_doubleclick_key_down == code &&
-            now - ictx->last_doubleclick_time < opts->doubleclick_time / 1000.0)
+            now - ictx->last_doubleclick_time < opts->doubleclick_time / 1000.0 &&
+            code >= MP_MBTN_LEFT && code <= MP_MBTN_RIGHT)
         {
-            if (code >= MP_MBTN_LEFT && code <= MP_MBTN_RIGHT) {
-                interpret_key(ictx, code - MP_MBTN_BASE + MP_MBTN_DBL_BASE,
-                              1, 1);
-            }
+            now = 0;
+            interpret_key(ictx, code - MP_MBTN_BASE + MP_MBTN_DBL_BASE,
+                          1, 1);
+        } else if (code == MP_MBTN_LEFT) {
+            // This is a mouse left botton down event which isn't part of a doubleclick.
+            // Initialize vo dragging in this case.
+            mp_cmd_t *cmd = mp_input_parse_cmd(ictx, bstr0("begin-vo-dragging"), "<internal>");
+            mp_input_queue_cmd(ictx, cmd);
         }
         ictx->last_doubleclick_key_down = code;
         ictx->last_doubleclick_time = now;
@@ -756,10 +762,12 @@ void mp_input_put_key(struct input_ctx *ictx, int code)
     input_unlock(ictx);
 }
 
-void mp_input_put_key_artificial(struct input_ctx *ictx, int code)
+void mp_input_put_key_artificial(struct input_ctx *ictx, int code, double value)
 {
+    if (value == 0.0)
+        return;
     input_lock(ictx);
-    mp_input_feed_key(ictx, code, 1, true);
+    mp_input_feed_key(ictx, code, value, true);
     input_unlock(ictx);
 }
 
@@ -1007,20 +1015,16 @@ void mp_input_enable_section(struct input_ctx *ictx, char *name, int flags)
 
     MP_TRACE(ictx, "enable section '%s'\n", name);
 
-    if (ictx->num_active_sections < MAX_ACTIVE_SECTIONS) {
-        int top = ictx->num_active_sections;
-        if (!(flags & MP_INPUT_ON_TOP)) {
-            // insert before the first top entry
-            for (top = 0; top < ictx->num_active_sections; top++) {
-                if (ictx->active_sections[top].flags & MP_INPUT_ON_TOP)
-                    break;
-            }
-            for (int n = ictx->num_active_sections; n > top; n--)
-                ictx->active_sections[n] = ictx->active_sections[n - 1];
+    int top = ictx->num_active_sections;
+    if (!(flags & MP_INPUT_ON_TOP)) {
+        // insert before the first top entry
+        for (top = 0; top < ictx->num_active_sections; top++) {
+            if (ictx->active_sections[top].flags & MP_INPUT_ON_TOP)
+                break;
         }
-        ictx->active_sections[top] = (struct active_section){name, flags};
-        ictx->num_active_sections++;
     }
+    MP_TARRAY_INSERT_AT(ictx, ictx->active_sections, ictx->num_active_sections,
+                        top, (struct active_section){name, flags});
 
     MP_TRACE(ictx, "active section stack:\n");
     for (int n = 0; n < ictx->num_active_sections; n++) {
@@ -1273,9 +1277,9 @@ static int parse_config(struct input_ctx *ictx, bool builtin, bstr data,
     return n_binds;
 }
 
-static int parse_config_file(struct input_ctx *ictx, char *file, bool warn)
+static bool parse_config_file(struct input_ctx *ictx, char *file)
 {
-    int r = 0;
+    bool r = false;
     void *tmp = talloc_new(NULL);
     stream_t *s = NULL;
 
@@ -1292,7 +1296,7 @@ static int parse_config_file(struct input_ctx *ictx, char *file, bool warn)
         MP_VERBOSE(ictx, "Parsing input config file %s\n", file);
         int num = parse_config(ictx, false, data, file, NULL);
         MP_VERBOSE(ictx, "Input config file %s parsed: %d binds\n", file, num);
-        r = 1;
+        r = true;
     } else {
         MP_ERR(ictx, "Error reading input config file %s\n", file);
     }
@@ -1317,6 +1321,7 @@ struct input_ctx *mp_input_init(struct mpv_global *global,
         .opts_cache = m_config_cache_alloc(ictx, global, &input_config),
         .wakeup_cb = wakeup_cb,
         .wakeup_ctx = wakeup_ctx,
+        .active_sections = talloc_array(ictx, struct active_section, 0),
     };
 
     ictx->opts = ictx->opts_cache->opts;
@@ -1373,13 +1378,13 @@ void mp_input_load_config(struct input_ctx *ictx)
 
     bool config_ok = false;
     if (ictx->opts->config_file && ictx->opts->config_file[0])
-        config_ok = parse_config_file(ictx, ictx->opts->config_file, true);
+        config_ok = parse_config_file(ictx, ictx->opts->config_file);
     if (!config_ok) {
         // Try global conf dir
         void *tmp = talloc_new(NULL);
         char **files = mp_find_all_config_files(tmp, ictx->global, "input.conf");
         for (int n = 0; files && files[n]; n++)
-            parse_config_file(ictx, files[n], false);
+            parse_config_file(ictx, files[n]);
         talloc_free(tmp);
     }
 
@@ -1392,6 +1397,14 @@ void mp_input_load_config(struct input_ctx *ictx)
     input_unlock(ictx);
 }
 
+bool mp_input_load_config_file(struct input_ctx *ictx, char *file)
+{
+    input_lock(ictx);
+    bool result = parse_config_file(ictx, file);
+    input_unlock(ictx);
+    return result;
+}
+
 static void clear_queue(struct cmd_queue *queue)
 {
     while (queue->first) {
